/*
  Part of the Processing Fullscreen API

  Copyright (c) 2006-08 Hansi Raber

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 3 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
package metasurface;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilePermission;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import processing.core.PApplet;

public class MetaSurface extends JFrame implements ActionListener, ChangeListener, CaretListener{
	private static final long serialVersionUID = 1L;

	String params[]; 
	Surface surface; 
	Surface.Entry activeEntry; 
	
	// int sliderResolution = 1000000000;
	int sliderResolution = Integer.MAX_VALUE; 
	
	// The current values
	public float[] values; 
	private float defaultValues[]; 
	
	private JSlider sliders[];
	private JLabel labels[]; 
	protected float[] minValues;
	protected float[] maxValues;
	
	private File currentFile; 
	
	private JTextField nameField;
	
	private ImageIcon openImg, saveImg, playImg, recordImg, stopImg, addImg, removeImg, renderMovieImg, renderImageImg, resetImg, warningImg; 
	protected JButton open, save, play, stop, record, stopRecord, sandboxWarning, renderImage;
	private JToggleButton renderMovie;
	
	private JButton addSet, removeSet, resetSet;
	
	protected PApplet sketch;
	
	public boolean hq; 
	int exportFrameRate = 25;
	private JToolBar toolbar;
	private boolean hasCustomEntries; 
	
	// Is there video loaded? 
	private boolean filesystemAvailable; 
	private boolean videoLibraryAvailable;
	
	public MetaSurface( PApplet sketch, String params[], float minValues[], float maxValues[], float values[] ){
		super( "Meta Surface" );
		setLayout( new BorderLayout() );  
		defaultValues = values.clone(); 
		
		this.sketch = sketch;
		this.values = values;  
		
		// test for some stuff...
		try{
			System.getSecurityManager().checkWrite( System.getProperty( "user.dir" ) );
			filesystemAvailable = true; 
		}
		catch( NullPointerException ex ){
			// no security manager! 
			filesystemAvailable = true; 
		}
		catch( SecurityException sex ){
			filesystemAvailable = false; 
		}
		
		try{
			Class.forName( "processing.video.MovieMaker" ); 
			videoLibraryAvailable = true; 
		}
		catch( ClassNotFoundException e ){
			System.err.println( "MetaSurface: Include the Video library to be able to record movies of your sketch. " ); 
			System.err.println( "             To do insert this line at the top of your sketch: " ); 
			System.err.println( ); 
			System.err.println( "import processing.video.*;" );
			System.err.println(); 
			System.err.println( "Original Error: " );
			e.printStackTrace();
		}
		catch( SecurityException sex ){
			System.err.println( "MetaSurface: The video library seems to be loaded, but can't be loaded. " );
			System.err.println( "             This happens because the sketch is running in a sandbox, " );
			System.err.println( "             propably as an applet. " );
			System.err.println(); 
			System.err.println( "Original Error: " );
			sex.printStackTrace();
		}
		catch( Throwable t ){
			// hm... very strange, let's better disable the video functionality! 
			System.err.println( "MetaSurface: The video library couldn't be loaded. " ); 
			System.err.println( "             No worries, we can live without it! " ); 
			System.err.println(); 
			System.err.println( "Original Error: " );
			t.printStackTrace();
		}
		
		// load images
		openImg = loadImage( "img/document-open.png" ); 
		saveImg = loadImage( "img/document-save.png" );
		playImg = loadImage( "img/media-playback-start.png" );
		stopImg = loadImage( "img/media-playback-stop.png" ); 
		recordImg = loadImage( "img/media-record.png" );
		addImg = loadImage( "img/list-add.png" ); 
		removeImg = loadImage( "img/list-remove.png" ); 
		renderMovieImg = loadImage( "img/camera-video.png" ); 
		renderImageImg = loadImage( "img/camera-photo.png" ); 
		resetImg = loadImage( "img/go-first.png" ); 
		warningImg = loadImage( "img/dialog-warning.png" ); 
		
		// Remember some of the variables passed along
		this.params = params;
		this.minValues = minValues;
		this.maxValues = maxValues;
		
		final JPanel content = new JPanel( new GridBagLayout() );
		content.setBorder( new EmptyBorder( 10, 10, 10, 10 ) ); 
		setContentPane( content ); 
		final GridBagConstraints gbc = new GridBagConstraints(); 
		
		gbc.gridx = 1; 
		gbc.weightx = 1; 
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.gridwidth = 2; 
		content.add( createButtons(), gbc ); 
		
		gbc.gridy = 1; 
		gbc.gridwidth = 1; 
		gbc.weightx = 1; 
		gbc.weighty = 1; 
		gbc.insets = new Insets( 5, 5, 5, 5 );
		gbc.fill = GridBagConstraints.BOTH; 
		gbc.anchor = GridBagConstraints.NORTHWEST; 
		content.add( createSurface(), gbc );
		
		gbc.gridx = 2; 
		gbc.weightx = 0; 
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.anchor = GridBagConstraints.NORTH; 
		content.add( createValueEditor(), gbc ); 
		
		updateSliders(); 
		
		// Remember size...
		getRootPane().putClientProperty("Window.style", "small"); 
		pack(); 
		//String key = sketch.getClass().toString();
		
		//MetaSurfaceHelper.rememberLocation( this, key, 100, 100 ); 
		//Helper.rememberSize( this, key, w, h ); 
		
		
		setResizable( false );
		setVisible( true ); 
	}
	
	public MetaSurface( PApplet sketch, String params[], float minValues[], float maxValues[] ){
		this( sketch, params, minValues, maxValues, minValues.clone() ); 
	}
	
	private Component createButtons() {
		toolbar = new JToolBar();
		toolbar.setFloatable( false );
		
		
		open = new JButton( "Open", openImg );
		save = new JButton( "Save", saveImg );
		play = new JButton( playImg );
		stop = new JButton( stopImg );
		record = new JButton( recordImg );
		stopRecord = new JButton( stopImg );
		renderMovie = new JToggleButton( renderMovieImg );
		renderImage = new JButton( renderImageImg ); 
		sandboxWarning = new JButton( "Restricted Mode", warningImg );
		
		open.setToolTipText( "Open a settings file" ); 
		save.setToolTipText( "Save a settings file" ); 
		play.setToolTipText( "Play back the recorded trail" ); 
		stop.setToolTipText( "Stop playback" ); 
		record.setToolTipText( "Start recording a trail" ); 
		stopRecord.setToolTipText( "Stop record a trail" ); 
		renderMovie.setToolTipText( "<html>Export the sketch to a movie file<br><i>Hold shift and click to start the trail playback synchroneously</i>" );
		renderImage.setToolTipText( "Export the sketch to an image file" ); 
		sandboxWarning.setToolTipText( "Warning: Running in restricted mode" );
		
		// use empty borders for the buttons
		open.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		save.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		play.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		stop.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		record.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		stopRecord.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) );
		renderMovie.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		renderImage.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		sandboxWarning.setBorder( BorderFactory.createEmptyBorder( 3, 3, 3, 3 ) ); 
		
		open.addActionListener( this ); 
		save.addActionListener( this );
		play.addActionListener( this ); 
		stop.addActionListener( this ); 
		record.addActionListener( this );
		stopRecord.addActionListener( this );
		renderMovie.addActionListener( this ); 
		renderImage.addActionListener( this ); 
		sandboxWarning.addActionListener( this ); 
		
		stop.setEnabled( false ); 
		stopRecord.setEnabled( false ); 
		
		if( filesystemAvailable ){
			toolbar.add( open ); 
			toolbar.add( save );
			toolbar.addSeparator(); 
		}
		
		toolbar.add( play ); 
		toolbar.add( stop ); 
		toolbar.addSeparator(); 
		toolbar.add( record ); 
		toolbar.add( stopRecord );
		
		if( filesystemAvailable ){
			toolbar.addSeparator(); 
			toolbar.add( renderImage );
			
			if( videoLibraryAvailable ){
				toolbar.add( renderMovie ); 
			}
		}
		
		
		if( !filesystemAvailable ){
			toolbar.addSeparator(); 
			toolbar.add( sandboxWarning ); 
		}
		
		return toolbar;
	}

	
	private JPanel createValueEditor() {
		JPanel panel = new JPanel( new GridBagLayout() );
		GridBagConstraints gbc = new GridBagConstraints(); 
		
		nameField = new JTextField( "unknown" );
		nameField.addCaretListener( this );  
		addComponent( panel, gbc, "Name: ", nameField, null ); 
		
		sliders = new JSlider[params.length]; 
		labels = new JLabel[params.length]; 
		for( int i = 0; i < params.length; i++ ){
			
			sliders[i] = new JSlider( 0, sliderResolution );
			sliders[i].putClientProperty("JComponent.sizeVariant", "small");
			sliders[i].addChangeListener( this );
			sliders[i].setPreferredSize( new Dimension( 250, sliders[i].getPreferredSize().height ) ); 
			labels[i] = new JLabel( "" + values[i] ); 
			labels[i].setPreferredSize( new Dimension( 75, labels[i].getPreferredSize().height ) );
			labels[i].setMinimumSize( labels[i].getPreferredSize() ); 
			labels[i].setFont( labels[i].getFont().deriveFont( 9f ) ); 
			addComponent( panel, gbc, params[i] + ": ", sliders[i], labels[i] ); 
		}
		// PApplet.println( values ); 
		
		// Create the toolbar for the sets
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable( false ); 
		addSet = new JButton( "Add as new", addImg ); 
		removeSet = new JButton( "Remove", removeImg ); 
		resetSet = new JButton( "Reset", resetImg );
		
		addSet.addActionListener( this ); 
		removeSet.addActionListener( this );
		resetSet.addActionListener( this ); 
		
		/*addSet.setBorderPainted( false ); 
		removeSet.setBorderPainted( false ); 
		resetSet.setBorderPainted( false );*/ 
		
		addSet.putClientProperty("JButton.buttonType", "gradient");  
		removeSet.putClientProperty("JButton.buttonType", "gradient");
		resetSet.putClientProperty("JButton.buttonType", "gradient");
		
		removeSet.setEnabled( false ); 
		
		toolbar.add( addSet ); 
		toolbar.add( removeSet ); 
		toolbar.add( resetSet ); 
		
		addComponent( panel, gbc, "", toolbar, null ); 
		
		/*JScrollPane scrollPane = new JScrollPane( panel ); 
		scrollPane.setHorizontalScrollBarPolicy( ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER ); 
		scrollPane.setVerticalScrollBarPolicy( ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED ); 
		scrollPane.setBorder( BorderFactory.createEmptyBorder() ); 
		return new JScrollPane( panel );*/
		return panel; 
	}
	
	private void addComponent( JPanel panel, GridBagConstraints gbc, String label, JComponent elem, JComponent elem2 ){
		gbc.gridy ++;
		gbc.gridx = 0; 
		gbc.weightx = 0; 
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridwidth = 1; 
		
		JLabel description = new JLabel( label );
		description.setFont( description.getFont().deriveFont( 9f ) ); 
		description.putClientProperty("JComponent.sizeVariant", "small");

		panel.add( description, gbc ); 
		
		gbc.gridx = 1; 
		gbc.weightx = 1;
		gbc.gridwidth = elem2 == null? 2:1; 
		panel.add( elem, gbc ); 
		
		if( elem2 != null ){
			gbc.weightx = 0; 
			gbc.gridx = 2; 
			panel.add( elem2, gbc ); 
		}
	}
	
	private JPanel createSurface() {
		JPanel container = new JPanel( new BorderLayout() );
		Border border = BorderFactory.createEtchedBorder(); 
		Insets insets = border.getBorderInsets( container ); 
		container.setBorder( border );
		
		surface = new Surface( this ); 
		Dimension dreamSize = surface.getPreferredSize();  
		container.setPreferredSize( new Dimension( dreamSize.width + insets.left + insets.right, dreamSize.height + insets.top + insets.bottom ) );
		
		surface.init();  
		container.add( surface );
		return container; 
	}
	
	
	private ImageIcon loadImage( String src ){
		try {
			return new ImageIcon( ImageIO.read( getClass().getResourceAsStream( src ) ) );
		} catch (Exception e) {
			e.printStackTrace(); 
			return null;  
		}
	}

	/**
	 * Action Performed. All the Buttons are handled here. 
	 * Awsome! 
	 */
	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource(); 
		
		// Start recording
		if( source == record ){
			surface.startRecording(); 
			record.setEnabled( false ); 
			stopRecord.setEnabled( true ); 
		}
		
		// Stop recording
		if( source == stopRecord ){
			surface.stopRecording();
			record.setEnabled( true ); 
			stopRecord.setEnabled( false ); 
		}
		
		// Start play
		if( source == play ){
			surface.startPlayback(); 
			play.setEnabled( false ); 
			stop.setEnabled( true ); 
		}
		
		// Stop play
		if( source == stop ){
			surface.stopPlayback(); 
			play.setEnabled( true ); 
			stop.setEnabled( false ); 
		}
		
		// Add an entry to the set
		if( source == addSet ){
			// Add a new entry
			float[] values = new float[sliders.length];
			for( int i = 0; i < sliders.length; i++ ){
				values[i] = getSliderValue( i );  
			}
			surface.addEntry( nameField.getText(), values ); 
		}
		
		// Remove an entry from the set 
		if( source == removeSet ){
			surface.removeEntry( activeEntry );
			activeEntry = null; 
		}
		
		// Reset to defaults
		if( source == resetSet ){
			for( int i = 0; i < sliders.length; i++ ){
				setSliderValue( i, defaultValues[i] ); 
			}
		}
		
		// Save the path to an xml file
		if( source == save ){
			if( currentFile == null || true ){
				String key = "metasurface.save." + sketch.getClass().getName(); 
				File fallback = new File( sketch.sketchPath ); 
				File newFile = MetaSurfaceHelper.showSaveDialog( key, currentFile, fallback, this, ".data" ); 
				
				if( newFile != null ){
					currentFile = newFile; 
					save( newFile ); 
				}
			}
		}
		
		// Load the set
		if( source == open ){
			String key = "metasurface.open." + sketch.getClass().getName(); 
			File fallback = new File( sketch.sketchPath );  
			File newFile = MetaSurfaceHelper.showOpenDialog( key, currentFile, fallback, this, ".data" );
			
			if( newFile != null ){
				currentFile = newFile; 
				open( newFile ); 
			}
		}
		
		// Export to movie file
		if( source == renderMovie ){
			if( renderMovie.isSelected() ){
				String key = "metasurface.renderVideo." + sketch.getClass().getName(); 
				File fallback = new File( sketch.sketchPath ); 
				File newFile = MetaSurfaceHelper.showSaveDialog( key, null, fallback, this, ".mov" ); 
				
				if( newFile == null ){
					renderMovie.setSelected( false ); 
				}
				else{
					surface.startRender( newFile );
					if( (e.getModifiers() & ActionEvent.SHIFT_MASK) != 0 ) surface.startPlayback(); 
				}
			}
			else{
				surface.stopRender();  
			}
		}
		
		// Render an image
		if( source == renderImage ){
			 String key = "metasurface.renderImage." + sketch.getClass().getName(); 
			 File fallback = new File( sketch.sketchPath ); 
			 File newFile = MetaSurfaceHelper.showSaveDialog( key, null, fallback, this, ".png,.tif,.tga,.jpg" ); 
			 
			 if( newFile == null ){
			 }
			 else{
				 sketch.saveFrame( newFile.getAbsolutePath() ); 
				 PApplet.open( newFile.getAbsolutePath() ); 
			 }
		}
		
		// Show the applet warning dialog
		if( source == sandboxWarning ){
			String warning =  
				"<html>" + 
				"This sketch is running with restricted permissions, "; 
			if( videoLibraryAvailable ){
				warning += 
					"therefor the following functionality in metasurface is disabled: <br>" +
					"<ul>" + 
					"	<li>Loading and saving of settings. </li>" + 
					"	<li>Exporting a sketch as movies or images</li>" + 
					"</ul>"; 
			}
			else{
				warning += "therefor loading and saving of settings is disabled. "; 
			}
			
			JOptionPane.showMessageDialog( this, warning, "Restricted Mode", JOptionPane.WARNING_MESSAGE ); 
		}
		
	}

	
	/**
	 * Load a path
	 * @param activeEntry
	 */
	public void save( File file ){
		try {
			surface.save( file );
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} 
	}
	
	/**
	 * Load a path
	 */
	public void open( File file ){
		try {
			surface.open( file );
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} 
	}
	
	/**
	 * Load from a stream
	 * @param activeEntry
	 */
	public void open( InputStream in ){
		try{
			surface.open( in ); 
		}
		catch( Exception e ){
			e.printStackTrace(); 
		}
	}
	
	public void setActiveEntry( Surface.Entry activeEntry ){
		this.activeEntry = activeEntry;
		
		if( activeEntry == null ){
			nameField.setText( "new" );
			removeSet.setEnabled( false ); 
		}
		else{
			nameField.setText( activeEntry.name );
			removeSet.setEnabled( true ); 
			
			System.arraycopy( activeEntry.values, 0, this.values, 0, this.values.length );
			updateSliders();
		}
	}

	public void updateSliders(){
		for( int i = 0; i < sliders.length; i++ ){
			setSliderValue( i, values[i] );		
		}
	}
	
	
	/**
	 * Convert from slider value to float value
	 */
	private float getSliderValue( int i ){
		return minValues[i] + sliders[i].getValue()*(maxValues[i]-minValues[i])*1f/sliderResolution;
	}
	
	/**
	 * Convert float value to slider value
	 */
	private int setSliderValue( int i, float value ){
		values[i] = value; 
		sliders[i].setValue( (int)((value - minValues[i])/(maxValues[i]-minValues[i])*sliderResolution) );
		labels[i].setText( "" + value ); 
		return sliders[i].getValue(); 
	}
	
	public void stateChanged( ChangeEvent e ){
		if( surface.isPlaying() ){
			return; 
		}
		
		for( int i = 0; i < sliders.length; i++ ){
			if( e.getSource() == sliders[i] ){
				values[i] = getSliderValue( i ); 
				labels[i].setText( "" + values[i] );
				if( activeEntry != null ){
					activeEntry.setValues( values );
					surface.pointerX = -10; 
					surface.pointerY = -10; 
				}
			}
		}
	}

	public void caretUpdate(CaretEvent e) {
		if( activeEntry != null ){
			activeEntry.setName( nameField.getText() ); 
		}
	}
	
	/**
	 * Register a listener for metasurface events. 
	 * To use this the object must have the method metaSurfaceEvent( String command ) defined. 
	 * For instance you can use it like this: 
	 * 
	 * <code>
	 * void setup(){
	 *   metaSurface = new MetaSurface( this, ... ); 
	 *   metaSurface.registerListener( this ); 
	 * }
	 * 
	 * void metaSurfaceEvent( String command ){
	 *   if( command.equals( "start" ) ){
	 *     println( "playback started" ); 
	 *   }
	 *   else if( command.equals( "stop" ){
	 *   	println( "playback stopped" ); 
	 *   }
	 * }
	 * </code>
	 * @param o
	 */
	public void registerListener( Object o ){
		surface.registerListener( o ); 
	}
	
	
	/**
	 * The frame rate at which movies are currently being exported.
	 *   
	 * @return The current export frame rate. 
	 */
	public int getExportFrameRate(){
		return exportFrameRate; 
	}
	
	/**
	 * Changes the framerate at which movies are exported. 
	 * Changing this during an export doesn't have any effect.  
	 * 
	 * @param frameRate The new frame rate
	 */
	public void setExportFrameRate( int frameRate ){
		this.exportFrameRate = frameRate; 
	}
	
	
	/**
	 * Adds a separator to the menu
	 */
	public void addSeparator(){
		toolbar.addSeparator(); 
	}
	
	/**
	 * Adds a button to the toolbar
	 * 
	 * Use like this: 
	 * 
	 * <code>
	 * void setup(){
	 *   ...
	 *   metaSurface = new MetaSurface( ... ); 
	 *   metaSurface.addButton( "my-icon.png", "buttonPressed" ); 
	 * }
	 * 
	 * void buttonPressed(){
	 *   println( "button was pressed" ); 
	 * }
	 * </code>
	 * 
	 * @param src The image source
	 * @param method The method to be called in your sketch. 
	 * @return The button created 
	 */
	public JButton addButton( String src, final String method ){
		JButton button; 
		// Try to load the image... 
		try{
			button = new JButton( new ImageIcon( ImageIO.read( sketch.createInput( src ) ) ) ); 
		}
		catch( Exception e ){
			button = new JButton( src ); 
		}
		
		button.setBorder( BorderFactory.createEmptyBorder() ); 
		button.addActionListener( new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				try{
					Method m = sketch.getClass().getMethod( method );
					m.invoke( sketch );
				}
				catch( Exception ex ){
					System.err.println( "MetaSurface: Couldn't invoke " + method );
				}
			};
		} ); 
		
		addCustomElement( button ); 
		
		return button; 
	}
	
	/**
	 * Adds a toggle button to the toolbar
	 * 
	 * Use like this: 
	 * 
	 * <code>
	 * void setup(){
	 *   ...
	 *   metaSurface = new MetaSurface( ... ); 
	 *   metaSurface.addButton( "my-icon.png", "buttonPressed" ); 
	 * }
	 * 
	 * void buttonPressed( boolean value ){
	 *   println( "button was pressed, value = " + value ); 
	 * }
	 * </code>
	 * 
	 * @param src The image source
	 * @param method The method to be called in your sketch. 
	 * @return The button created 
	 */
	public JToggleButton addToggleButton( String src, final String method ){
		JToggleButton button; 
		// Try to load the image... 
		try{
			button = new JToggleButton( new ImageIcon( ImageIO.read( sketch.createInput( src ) ) ) ); 
		}
		catch( Exception e ){
			button = new JToggleButton( src ); 
		}
		
		button.addActionListener( new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				try{
					Method m = sketch.getClass().getMethod( method, boolean.class );
					m.invoke( sketch, ((JToggleButton)e.getSource()).isSelected() ); 
				}
				catch( Exception ex ){
					System.err.println( "MetaSurface: Couldn't invoke " + method );
					ex.printStackTrace(); 
				}
			};
		} ); 
		
		addCustomElement( button ); 
		
		return button; 
	}
	
	/**
	 * Adds a select field to the toolbar
	 * 
	 * Use like this: 
	 * 
	 * <code>
	 * void setup(){
	 *   ...
	 *   metaSurface = new MetaSurface( ... ); 
	 *   metaSurface.addSelectField( "Red", "Green", "Blue", "valueChanged" ); 
	 * }
	 * 
	 * void valueChanged( String value ){
	 *   println( "value = " + value ); 
	 * }
	 * </code>
	 * 
	 * @param values The different values, the last one being the function that is called when the value changes. 
	 * @return The button created 
	 */
	public JComboBox addSelectList( String... values ){
		final JComboBox list = new JComboBox( );
		list.setEditable( false ); 
		
		for( int i = 0; i < values.length - 1; i++ ){
			list.addItem( values[i] );
		}
		
		final String method = values[values.length-1]; 
		
		list.addActionListener( new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				try{
					Method m = sketch.getClass().getMethod( method, String.class );
					m.invoke( sketch, list.getSelectedItem().toString() ); 
				}
				catch( Exception ex ){
					System.err.println( "MetaSurface: Couldn't invoke " + method );
					ex.printStackTrace(); 
				}
			};
		} ); 
		
		return list; 
	}
	
	/**
	 * Add a custom element to the toolbar
	 */
	public void addCustomElement( JComponent component ){
		if( !hasCustomEntries ){
			hasCustomEntries = true; 
			// toolbar.addSeparator(); 
		}
		
		toolbar.add( component ); 
	}

	
	/**
	 * Starts playback
	 */
	public void startPlayback(){
		surface.startPlayback(); 
	}
	
	/**
	 * Stop playback
	 */
	public void stopPlayback(){
		surface.stopPlayback(); 
	}
}
